use nom::combinator::recognize;
use nom::error::VerboseError;
use nom::IResult;

use super::text::bold_end;
use super::text::Res;

type Matcher = dyn for<'s> Fn(&'s str) -> IResult<&'s str, &'s str, VerboseError<&'s str>>;
type Link<'r, T> = Option<&'r Node<'r, T>>;

pub trait OrgModeContext<'r> {
    // todo
}

struct FailMatcherNode<'r> {
    fail_matcher: ChainBehavior<'r>,
}

struct PreviousElementNode<'r> {
    dummy: &'r str,
}

enum ChainBehavior<'r> {
    AndParent(Option<&'r Matcher>),
    IgnoreParent(Option<&'r Matcher>),
}

struct Node<'r, T: 'r + ?Sized> {
    elem: &'r T,
    parent: Link<'r, T>,
}

struct ContextTree<'r, T: 'r + ?Sized> {
    // Not using Link so the ContextTree can own this node
    head: Option<Node<'r, T>>,
}

impl<'r> ContextTree<'r, dyn OrgModeContext<'r>> {
    pub fn new() -> Self {
        ContextTree { head: None }
    }

    pub fn with_additional_node(
        &'r self,
        new_elem: &'r dyn OrgModeContext<'r>,
    ) -> ContextTree<'r, dyn OrgModeContext<'r>> {
        let new_node = Node {
            elem: new_elem,
            parent: self.head.as_ref(),
        };
        ContextTree {
            head: Some(new_node),
        }
    }
}

impl<'r> OrgModeContext<'r> for FailMatcherNode<'r> {}
impl<'r> OrgModeContext<'r> for PreviousElementNode<'r> {}

fn recognize_bold_end(input: &str) -> Res<&str, &str> {
    recognize(bold_end)(input)
}

fn test_context() {
    let foo = "foo";
    let context = ContextTree::new();
    let child1_context = PreviousElementNode { dummy: foo };
    let child1 = context.with_additional_node(&child1_context);
    let child2_context = FailMatcherNode {
        fail_matcher: ChainBehavior::AndParent(Some(&recognize_bold_end)),
    };
    let child2 = child1.with_additional_node(&child2_context);
}
